"""
note:
    pytermgui color names for basic colors:
        there are some name changes between 4.1.0 and 4.2.0, for example the
        color number 8 before is 'grey', after is 'bright-black', number 9
        before is 'brightred', after is 'bright-red'.
        i'd like to keep pace with the latest version of ptg. so i'll use
        'bright-black', not 8 although it's not quite compatible insurance.
        *this decision remains to be changeable in the future.*
        ref: `pytermgui/colors.py > (var) XTERM_NAMED_COLORS`
    there are some markup issues should be known. see `gitbook > ptg-markup.md`.
"""
import re
from textwrap import indent
from typing import Union

from ._internal_debug import debug  # noqa


class MessageFormatter:
    # see `self.fmt_message`.
    _braket_pattern = re.compile(r'\[[-/@_a-z! ():]+]')
    
    def fmt_source(self, filepath: str, lineno: Union[int, str],
                   is_external_lib: bool = False, fmt_width=False) -> str:
        if fmt_width:
            text_a = f'{filepath}:{lineno}'
            text_b = self._fmt_width(text_a, min_width=16)
            additional_space = ' ' * (len(text_b) - len(text_a))
        else:
            additional_space = ''
        if is_external_lib:
            assert filepath.startswith('[')
            a, b = filepath[1:].split(']', 1)
            filepath = (
                '[bold {color}]\\[{a}]'
                '[bold blue]{b}[/]'.format(
                    a=a, b=b, color='red' if filepath.startswith('[unknown]')
                    else 'magenta'
                )
            )
            # FIXME: `tim.print('[blue][magenta]aaa[blue]bbb')`
            #   'bbb' color is invalid!
            #   below is a workaround.
            return (
                '{0}'
                '[bright-black]:[/]'
                '[blue]{1}{2}[/]'.format(
                    filepath, lineno, additional_space
                )
            )
        return (
            '[bold blue]{0}[/]'
            '[bright-black]:[/]'
            '[blue]{1}{2}[/]'.format(
                filepath, lineno, additional_space
            )
        )
    
    @staticmethod
    def fmt_separator(sep: str = ' >> ') -> str:
        return '[bright-black]{}[/]'.format(sep)
    
    def fmt_funcname(self, funcname: str, fmt_width=False) -> str:
        is_func = not funcname.startswith('<')
        if is_func:
            funcname += '()'
        else:
            funcname = funcname[1:-1]
        if fmt_width:
            funcname = self._fmt_width(
                funcname, min_width=8, incremental_unit=2
            )
        if is_func:
            return '[green]{}[/]'.format(funcname)
        else:
            return '[magenta]{}[/]'.format(funcname)
    
    @staticmethod
    def fmt_index(idx: Union[int, str]) -> str:
        return '[red]\\[{}][/]'.format(idx)
    
    @staticmethod
    def fmt_divider(div_: str) -> str:
        return '[yellow]{}[/]'.format(div_)
    
    def fmt_message(self, msg: str, rich: bool, multilines=False) -> str:
        if rich:
            if multilines:
                x = []
                for line in msg.splitlines():
                    if line.startswith('    '):
                        how_many = len(line) - len(line.lstrip())
                        spaces = line[:how_many]
                        rich_indent = spaces.replace(
                            '    ', '[bright-black]│[/]   '
                        )
                        x.append(rich_indent + line[how_many:])
                    else:
                        x.append(line)
                out = '\n' + indent('\n'.join(x), '    ')
                return out
            else:
                return msg.replace(';\t', '[bright-black];\t[/]')
        else:
            # FIXME: this is a workaround for pytermgui's parser.
            #   1. ptg cannot parse backslash, so we convert it to slash.
            #   2. ptg tries to parse any valid patterns, but ignore the
            #      invalid ones. we should add a backslash to the former, but
            #      do nothing to the latter.
            msg = msg.replace('\\', '■')
            # # msg = msg.replace('[', '\\[')
            msg = self._braket_pattern.sub(lambda x: '\\' + x.group(), msg)
            if '■' in msg:
                # # msg = msg.replace('■', '[magenta]/[/]')
                msg = re.sub('■+', lambda x: r'[magenta]{}[/]'.format(
                    '/' * len(x.group())), msg)
            # debug(f'{msg = }')
            if multilines:
                out = '\n' + indent(msg, '    ')
                return out
            else:
                return msg.replace(';\t', '[bright-black];\t[/]')
    
    @staticmethod
    def fmt_level(text: str, level: str) -> str:
        colors = {
            'trace': '',
            'debug': 'bright-black',
            'info' : 'bright-blue',
            'warn' : 'yellow',
            'error': 'bright-red',
            'fatal': 'bold bright-cyan @bright-red',
        }
        if c := colors.get(level):
            # debug('[{}]{}[/]'.format(c, text))
            return '[{}]{}[/fg]'.format(c, text)
        else:
            return text
    
    _width_cache = {}  # dict[int raw_width_of_text, int suggested_width]
    
    def _fmt_width(self, text: str,
                   min_width: int = None,
                   incremental_unit=4) -> str:
        
        def _fmt(text, suggested_width) -> str:
            return '{{:<{}}}'.format(suggested_width).format(text)
        
        if len(text) in self._width_cache:
            return _fmt(text, self._width_cache[len(text)])
        
        if min_width is None:
            min_width = incremental_unit
        if len(text) <= min_width:
            suggested_width = min_width
        else:
            increasing_part = len(text) - min_width
            if increasing_part % incremental_unit == 0:
                suggested_width = min_width + increasing_part
            else:
                suggested_width = min_width + (
                        increasing_part // incremental_unit + 1
                ) * incremental_unit
        self._width_cache[len(text)] = suggested_width
        return _fmt(text, suggested_width)



import os
from inspect import currentframe

from pytermgui import parser
from pytermgui import tim
from pytermgui.prettifiers import prettify

from ._internal_debug import debug  # noqa
from .general import normpath
from .general import std_print
from .markup import MarkupAnalyser
from .message_formatter import MessageFormatter
from .path_helper import PathHelper
from .sourcemap import sourcemap

setattr(parser, 'print', std_print)
_formatter = MessageFormatter()


class LoggingConfig:
    """
    options:
        show_source: bool[true]
            add source info (filepath and line number) prefix to log messages.
            example:
                lk.log('hello world')
                # enabled : './main.py:10  >>  hello world'
                # disabled: 'hello world'
        show_varnames: bool[false]
            show both var names and values. (magic reflection)
            example:
                a, b = 1, 2
                lk.log(a, b, a + b)
                # enabled : 'main.py:11  >>  a = 1; b = 2; a + b = 3'
                # disabled: 'main.py:11  >>  1, 2, 3'
        show_external_lib: bool[true]
            if `param source` came from an external library, whether to print.
            for example, if a third-party library 'xxx' also used `lk.log`,
            its source path (relative to current working dir) may be very long,
            if you don't want to see any prints except your own project, you'd
            set this to False.
        
        # the following options are only available if `show_external_lib` is
        # true.
        path_format_for_external_lib: literal
            literal:
                'pretty_relpath': default
                    trunscate the source path of external lib to be shorter.
                    example:
                        before:
                            '../../../../site-packages/lk_logger/sourcemap.py'
                            # there may be a lot of '../'.
                        after:
                            '[lk_logger]/sourcemap.py'
                'relpath':
                    a relative path to current working dir. (<- `os.getcwd()`)
                    note there may be a lot of '../../../...' if external lib
                    is far beyond the current working dir.
                'lib_name_only':
                    show only the library name (surrounded by brackets).
                    example: '[lk_logger]'
            ps: if you don't want to show anything, you should turn to set
            `show_external_lib` to False.
    """
    show_source = True
    show_varnames = False
    show_external_lib = True
    path_format_for_external_lib = 'pretty_relpath'
    
    def update(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)


class LKLogger:
    
    def __init__(self):
        # self._cache = {}  # TODO: be complete in next version
        # #   dict[tuple k, namedtuple v]
        # #       k: (filepath, lineno)
        # #       v: ...
        self._config = LoggingConfig()
        self._counter = 0
        self._cwd = normpath(os.getcwd())
        self._markup_analyser = MarkupAnalyser()
        
        # lazy loaded
        self.__external_libs = None
        self.__proj_root = None
    
    def configure(self, **kwargs):
        self._config.update(**kwargs)
    
    @property
    def _proj_root(self) -> str:
        if self.__proj_root is None:
            self.__proj_root = PathHelper.find_project_root()
        return self.__proj_root
    
    @property
    def _external_libs(self) -> dict:
        if self.__external_libs is None:
            self.__external_libs = PathHelper.indexing_external_libs()
        return self.__external_libs
    
    def log(self, *args, **_):
        msg = self._main(currentframe().f_back, *args)
        # debug(msg)
        tim.print(msg)
    
    def fmt(self, *args, **_):
        return self._main(currentframe().f_back, *args)
    
    # noinspection PyUnresolvedReferences
    def _main(self, frame, *args) -> str:
        message_details = {
            'divider_line': '',
            'filepath'    : '',
            'funcname'    : '',
            'index'       : '',
            'lineno'      : '0',
            'log_level'   : '',
            'message'     : '',
        }
        
        is_external_lib = False
        marks = {}
        markup_pos = 0  # 0 not exist, 1 for begining, -1 for ending.
        traceback_level = 0  # int[default 1]
        
        if args:
            if (
                    isinstance(args[0], str) and args[0].startswith(':')
            ):
                markup_pos = 1
                markup, args = args[0], args[1:]
            elif len(args) > 1 and (
                    isinstance(args[-1], str) and args[-1].startswith(':')
            ):
                markup_pos = -1
                markup, args = args[-1], args[:-1]
            else:
                markup_pos = 0
                markup, args = None, args
            
            # analyse markup
            if markup:
                # analyse markup
                marks = self._markup_analyser.analyse_markup(markup)
                if marks:
                    if 'd' in marks:
                        message_details['divider_line'] = '-' * 80
                    if 'i' in marks:
                        if marks['i'] == 0:
                            self._counter = 0
                            if args:
                                self._counter += 1
                            else:
                                message_details['message'] = \
                                    '[bright-black](index reset)[/]'
                                # make sure 'r' key exists.
                                marks.setdefault('r', 0)
                        else:
                            self._counter += 1
                        message_details['index'] = str(self._counter)
                    if 'l' in marks:
                        args = tuple((
                            prettify(x, indent=4)
                            for x in args
                        ))
                    if 'p' in marks:
                        traceback_level = marks['p']
                    if 'r' in marks:
                        pass
                    if 's' in marks:
                        pass
                    if 'v' in marks:
                        message_details['log_level'] = (
                            'trace', 'debug', 'info',
                            'warn', 'error', 'fatal'
                        )[marks['v']]
        
        info = sourcemap.get_sourcemap(
            frame=frame,
            traceback_level=traceback_level,
            advanced=self._config.show_varnames,
        )
        
        if args:
            if self._config.show_varnames:
                if markup_pos == 0:
                    varnames = info.varnames
                elif markup_pos == 1:
                    varnames = info.varnames[1:]
                else:
                    varnames = info.varnames[:-1]
                try:
                    assert len(varnames) == len(args), (varnames, args)
                except AssertionError:
                    # debug('failed extracting varnames')
                    message_details['message'] = ';\t'.join(map(str, args))
                else:
                    # organize args
                    tmp_msg = []
                    for n, a in zip(varnames, args):
                        tmp_msg.append(f'{n} = {a}' if n else str(a))
                    message_details['message'] = ';\t'.join(tmp_msg)
            else:
                message_details['message'] = ';\t'.join(map(str, args))
        
        message_details['funcname'] = info.funcname
        
        # show external lib?
        if self._config.show_external_lib:
            # is external lib?
            if self._is_external_lib(info.filepath):  # yes
                is_external_lib = True
                # path format
                fmt = self._config.path_format_for_external_lib
                if fmt == 'relpath':
                    message_details['filepath'] = normpath(
                        os.path.relpath(info.filepath, self._cwd)
                    )
                else:
                    for lib_path in sorted(self._external_libs, reverse=True):
                        if info.filepath.startswith(lib_path):
                            lib_name = self._external_libs[lib_path].lower()
                            lib_relpath = \
                                info.filepath[len(lib_path):].lstrip('./') or \
                                os.path.basename(info.filepath)
                            break
                    else:
                        lib_name = ''
                        lib_relpath = info.filepath.lstrip('./')
                    
                    # debug(lib_name, lib_relpath)
                    
                    if fmt == 'pretty_relpath':
                        if lib_name:
                            message_details['filepath'] = \
                                '[{}]/{}'.format(lib_name, lib_relpath)
                        else:
                            message_details['filepath'] = \
                                '[{}]/{}'.format('unknown', lib_relpath)
                    elif fmt == 'lib_name_only':
                        message_details['filepath'] = f'[{lib_name}]'
            else:  # no
                message_details['filepath'] = normpath(
                    os.path.relpath(info.filepath, self._cwd)
                )
        else:
            if self._is_external_lib(info.filepath):
                pass  # is_external_lib = True ?
            else:
                message_details['filepath'] = normpath(
                    os.path.relpath(info.filepath, self._cwd)
                )
        if (x := message_details['filepath']) and \
                not x.startswith(('../', '[')):
            message_details['filepath'] = './' + x
        
        message_details['lineno'] = str(info.lineno)
        
        # show message
        global _formatter
        # noinspection PyListCreation
        message_elements = []
        message_elements.append(
            _formatter.fmt_source(
                message_details['filepath'],
                message_details['lineno'],
                is_external_lib=is_external_lib,
                fmt_width=True
            )
        )
        message_elements.append(
            _formatter.fmt_separator('\t>>\t')
        )
        message_elements.append(
            _formatter.fmt_funcname(
                message_details['funcname'],
                fmt_width=True
            )
        )
        message_elements.append(
            _formatter.fmt_separator('\t>>\t')
        )
        if message_details['log_level']:
            message_elements.append(
                _formatter.fmt_level(
                    text='[bold][{}][/bold]'.format(
                        message_details['log_level'].upper()
                    ), level=message_details['log_level']
                )
            )
            message_elements.append('\t')
            #   why use '\t', not ' '?
            #   '\t' aligns the following message tidier.
            #   comparison:
            #       use ' ':            use '\t':
            #           [TRACE] ...         [TRACE] ...
            #           [DEBUG] ...         [DEBUG] ...
            #           [INFO] ...          [INFO]  ...
            #           [WARN] ...          [WARN]  ...
            #                 ^                   ^^
            '''
            bug: ptg consecutive duplicate markup invalid:
                tim.print('[yellow]aaa[/][red]bbb[/]')
                tim.print('[yellow]aaa[/][yellow]bbb[/]')  # <- bug
                                      ^^^ B     ^^^ A
                tim.print('[yellow]aaa[/] [yellow]bbb[/]')
            part `A` doesn't rendered as yellow because its previous mark is
            also yellow. we have to remove `B` to make it work.
            the following code is a WORKAROUND!
            '''
            if message_details['log_level'] == 'debug':
                message_elements[-3] = message_elements[-3][:-3]
        if message_details['index']:
            message_elements.append(
                _formatter.fmt_index(
                    message_details['index']
                )
            )
            message_elements.append(' ')
        if message_details['divider_line']:
            message_elements.append(
                _formatter.fmt_divider(
                    message_details['divider_line']
                )
            )
            message_elements.append(' ')
        if 'l' in marks:
            message_details['message'] = message_details['message'].replace(
                ';\t', '\n'
            )
        message_elements.append(
            _formatter.fmt_message(
                message_details['message'],
                rich='r' in marks or 'l' in marks,
                multilines='l' in marks,
            )
        )
        if message_details['log_level'] and \
                message_details['log_level'] != 'trace':
            message_elements[-1] = _formatter.fmt_level(
                message_elements[-1],
                level=message_details['log_level']
            )
        
        return ''.join(message_elements)
    
    def _is_external_lib(self, filepath: str) -> bool:
        return not filepath.startswith(self._proj_root)


lk = LKLogger()



import os
import sys

from .general import normpath


class PathHelper:
    @staticmethod
    def find_project_root() -> str:
        """ find personal-like project root.

        proposals:
            1. backtrack from current working dir to find a folder which has
               one of '.idea', '.git', etc. folders.
            2. iterate paths in sys.path, to find a folder which is parent of
               current working dir. (adopted)
               if there are more than one, choose which is shortest.
               if there is none, return current working dir.
        """
        cwd = normpath(os.getcwd())
        paths = tuple(
            x for x in map(normpath, map(os.path.abspath, sys.path))
            if cwd.startswith(x) and os.path.isdir(x)
        )
        if len(paths) == 0:
            return cwd
        elif len(paths) == 1:
            return paths[0]
        else:
            return min(paths, key=lambda x: len(x))
    
    @staticmethod
    def indexing_external_libs() -> dict:
        """
        return:
            dict[str path, str lib_name]
        """
        out = {}
        
        for path in reversed(sys.path):
            if not os.path.exists(path):
                continue
            for root, dirs, files in os.walk(path):
                root = normpath(root)
                out[root] = os.path.basename(root)
                
                for d in dirs:
                    if d.startswith(('.', '__')):
                        continue
                    if '-' in d or '.' in d:
                        continue
                    out[f'{root}/{d}'] = d
                
                # for f in files:
                #     name, ext = os.path.splitext(f)
                #     if ext not in ('.py', '.pyc', '.pyd', '.pyo', '.pyw'):
                #         continue
                #     if '-' in name or '.' in name:
                #         continue
                #     out[f'{root}/{f}'] = name
                break
        
        return out



import builtins

from .general import std_print
from .logger import lk


def setup(**kwargs):
    """
    args:
        kwargs: see `./logger.py > LoggingConfig`.
    """
    quiet = kwargs.pop('quiet', False)
    
    lk.configure(**kwargs)
    setattr(builtins, 'print', lk.log)
    
    if not quiet:
        from random import randint
        slogan = (
            '[!rainbow]lk-logger is ready[/!rainbow]',
            '[!gradient(197)]lk-logger is ready[/!gradient]',
            '[!gradient(197)]lk-logger[/!gradient] [171]is ready[/]',
        )[randint(0, 2)]
        print('[bold red]♥[/] ' + slogan + ' [bold red]♥[/]', ':rp')


def unload():
    setattr(builtins, 'print', std_print)


def enable():
    setattr(builtins, 'print', lk.log)


def disable():
    setattr(builtins, 'print', lambda *_, **__: None)



"""
TODO: WIP (for 5.0.2)
"""
from pytermgui.parser import RE_MARKUP
import re


class RichText:
    _ptg_mark_pattern = re.compile(r'\[[-/@_a-z! ():]+]')
    _text: str
    
    def __init__(self):
        self._text = ''
    
    def __add__(self, other):
        self._text += other
        return self
    
    def __radd__(self, other):
        self._text = other + self._text
        return self
    
    def __str__(self) -> str:
        out = ''
        for seg in self._text.split('['):
            seg = '[' + seg
            if self._ptg_mark_pattern.match(seg):
                pass



class MarkupAnalyser:
    """
    readme:
        ~/docs/markup.zh.md

    markup list:
        :d  divider line
        :i  index
        :l  long / loose
        :p  parent layer
        :r  rich
        :s  short
        :t  tag
        :v  verbose / log level
    """
    
    def __init__(self):
        from re import compile
        self._mark_pattern = compile(r'[a-z]\d*')
    
    def analyse_markup(self, markup: str) -> dict:
        """
        return:
            dict[literal mark, int value]
                mark: d, i, l, p, r, s, t, v
        """
        out = {}
        defaults = {
            'd': 0,
            'i': 1,
            'l': 0,
            'p': 1,
            'r': 0,
            's': 0,
            't': 0,
            'v': 1,
        }
        marks = self._mark_pattern.findall(markup) or ()  # list[str]
        for m in marks:
            if len(m) == 1:
                out[m] = defaults[m]
            else:
                out[m[0]] = int(m[1:])
        return out
